@@ -33,6 +33,7 @@
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/stat.h>
+#include <sys/ioctl.h>
 #include <dirent.h>
 #include <ctype.h>
 #include <errno.h>
@@ -41,6 +42,7 @@
 #include <float.h>
 #include <fcntl.h>
 #include <limits.h>
+#include <ncurses.h>
 
 #define APP_NAME		"cpustat"
 #define TABLE_SIZE		(2411)		/* Should be a prime */
@@ -60,6 +62,7 @@
 #define OPT_SAMPLES		(0x00000800)
 #define OPT_DISTRIBUTION	(0x00001000)
 #define OPT_EXTRA_STATS		(0x00002000)
+#define OPT_TOP			(0x00004000)
 
 /* Histogram specific constants */
 #define MAX_DIVISIONS		(20)
@@ -188,6 +191,18 @@ typedef struct {
 	uint16_t offset;
 } proc_stat_fields_t;
 
+typedef struct {
+	void (*df_setup)(void);
+	void (*df_endwin)(void);
+	void (*df_clear)(void);
+	void (*df_refresh)(void);
+	void (*df_winsize)(bool redo);
+	void (*df_printfnl)(const char *fmt, ...);
+	void (*df_putc)(int ch);
+	void (*df_putstr)(char *str, int n);
+	void (*df_linebreak)(void);
+} display_funcs_t;
+
 static cpu_freq_scale_t cpu_freq_scale[] = {
 	{ 1e1,  1e0,  "Hz" },
 	{ 1e4,  1e3,  "KHz" },
@@ -234,6 +249,12 @@ static unsigned int opt_flags;		/* option flags */
 static uint64_t clock_ticks;		/* number of clock ticks per second */
 static pid_t opt_pid = -1;		/* PID to match against, -p option */
 
+static bool resized;			/* window resized */
+static int rows = 25;			/* tty size, rows */
+static int cols = 80;			/* tty size, columns */
+
+static display_funcs_t df;		/* display functions */
+
 /*
  *  Attempt to catch a range of signals so
  *  we can clean
@@ -287,6 +308,199 @@ static const int signals[] = {
 	-1,
 };
 
+/*
+ *  handle_sigwinch()
+ *      flag window resize on SIGWINCH
+ */
+static void handle_sigwinch(int sig)
+{
+	(void)sig;
+
+	df.df_winsize(true);
+	resized = true;
+}
+
+/*
+ *   cpustat_top_setup
+ *	setup display in top mode
+ */
+static void cpustat_top_setup(void)
+{
+	initscr();
+	cbreak();
+	noecho();
+	nodelay(stdscr, 1);
+	keypad(stdscr, 1);
+	curs_set(0);
+}
+
+/*
+ *  cpustat_generic_winsize()
+ *	get tty size in top mode
+ */
+static void cpustat_generic_winsize(bool redo)
+{
+	if (redo) {
+		struct winsize ws;
+
+		if (ioctl(fileno(stdin), TIOCGWINSZ, &ws) != -1) {
+			rows = ws.ws_row;
+			cols = ws.ws_col;
+		} else {
+			rows = 25;
+			cols = 80;
+		}
+	}
+}
+
+static void cpustat_top_winsize(bool redo)
+{
+	(void)redo;
+
+	cpustat_generic_winsize(true);
+	resizeterm(rows, cols);
+}
+
+/*
+ *  cpustat_noop()
+ *	no-operation void handler
+ */
+static void cpustat_noop(void)
+{
+}
+
+
+/*
+ *  cpustat_top_endwin()
+ *	call endwin in top mode
+ */
+static void cpustat_top_endwin(void)
+{
+	df.df_winsize(true);
+	resizeterm(rows, cols);
+	refresh();
+	resized = false;
+	clear();
+	endwin();
+}
+
+/*
+ * cpustat_top_clear()
+ *	clear screen in top mode
+ */
+static inline void cpustat_top_clear(void)
+{
+	clear();
+}
+
+/*
+ *  cpustat_top_refresh()
+ *	refresh screen in top mode
+ */
+static inline void cpustat_top_refresh(void)
+{
+	refresh();
+}
+
+/*
+ * cpustat_generic_printfnl()
+ *      cpustat generic printf with newline
+ */
+static void cpustat_generic_printfnl(const char *fmt, ...)
+{
+	va_list ap;
+	int n;
+
+	va_start(ap, fmt);
+	char buf[256];
+
+	n = vsnprintf(buf, sizeof(buf) - 2, fmt, ap);
+	df.df_putstr(buf, n);
+	va_end(ap);
+}
+
+/*
+ *  cpustat_top_putc()
+ *	cpustat put char in top mode
+ */
+static void cpustat_top_putc(int ch)
+{
+	addch(ch);
+}
+
+/*
+ *  cpustat_nomnal_putc()
+ *	cpustat put char in normal mode
+ */
+static void cpustat_normal_putc(int ch)
+{
+	fputc(ch, stdout);
+}
+
+/*
+ *  cpustat_top_putstr)
+ * 	cpustat put string in top mode
+ */
+static void cpustat_top_putstr(char *str, int n)
+{
+	if (n > cols) {
+		n = cols;
+		str[n] = '\0';
+	} else {
+		str[n] = '\n';
+		str[n + 1] = '\0';
+	}
+
+	addstr(str);
+}
+
+/*
+ *  cpustat_normal_putstr()
+ * 	cpustat put string in normal mode
+ */
+static void cpustat_normal_putstr(char *str, int n)
+{
+	if (n > cols)
+		n = cols;
+
+	str[n] = '\n';
+	str[n + 1] = '\0';
+
+	fputs(str, stdout);
+}
+
+/*
+ *  cpustat_normal_linebreak()
+ *	put a line break between output
+ */
+static void cpustat_normal_linebreak(void)
+{
+	putc('\n', stdout);
+}
+
+static display_funcs_t df_top = {
+	cpustat_top_setup,
+	cpustat_top_endwin,
+	cpustat_top_clear,
+	cpustat_top_refresh,
+	cpustat_top_winsize,
+	cpustat_generic_printfnl,
+	cpustat_top_putc,
+	cpustat_top_putstr,
+	cpustat_noop,
+};
+
+static display_funcs_t df_normal = {
+	cpustat_noop,
+	cpustat_noop,
+	cpustat_noop,
+	cpustat_noop,
+	cpustat_generic_winsize,
+	cpustat_generic_printfnl,
+	cpustat_normal_putc,
+	cpustat_normal_putstr,
+	cpustat_normal_linebreak,
+};
 
 /*
  *  strtouint64()
@@ -302,7 +516,7 @@ static uint64_t OPTIMIZE3 HOT strtouint64(char *str, char **endptr)
 
 		if (digit > 9)
 			break;
-		if (UNLIKELY(v >= 1844674407370955161))
+		if (UNLIKELY(v >= 1844674407370955161ULL))
 			goto do_overflow;
 		v *= 10;
 		v += digit;
@@ -521,7 +735,7 @@ no_cmd:
 		bool new_info = false;
 		/*
 		 * We may be re-using a stale old PID, or we may need
-		 * a new info struct 
+		 * a new info struct
 		 */
 		if (!info) {
 			info = malloc(sizeof(pid_info_t));
@@ -686,14 +900,14 @@ static inline double duration_round(const double duration)
         return floor((duration * 100.0) + 0.5) / 100.0;
 }
 
-static inline void putdec(const int v)
+static inline void putdec(char *str, const int v)
 {
 	register int d;
 
 	d = (v / 10);
-	fputc(d > 9 ? '?' : '0' + d, stdout);
-	d = (v % 10);
-	fputc('0' + d, stdout);
+	str[0] = d > 9 ? '?' : '0' + d;
+	d = v % 10;
+	str[1] = '0' + d;
 }
 
 /*
@@ -702,20 +916,30 @@ static inline void putdec(const int v)
  */
 static void info_banner_dump(const double time_now)
 {
-	fputs("  %CPU   %USR   %SYS   PID S  CPU   Time Task", stdout);
+	char str[256] = "  %CPU   %USR   %SYS   PID S  CPU   Time Task";
+	char *ptr = str + strlen(str);
+
 	if (opt_flags & OPT_TIMESTAMP) {
 		struct tm tm;
 
 		get_tm(time_now, &tm);
-		fputs("  (", stdout);
-		putdec(tm.tm_hour);
-		fputc(':', stdout);
-		putdec(tm.tm_min);
-		fputc(':', stdout);
-		putdec(tm.tm_sec);
-		fputc(')', stdout);
+		strncpy(ptr, "  (", 3);
+		ptr += 3;
+		putdec(ptr, tm.tm_hour);
+		ptr += 2;
+		*ptr = ':';
+		ptr++;
+		putdec(ptr, tm.tm_min);
+		ptr += 2;
+		*ptr = ':';
+		ptr++;
+		putdec(ptr, tm.tm_sec);
+		ptr += 2;
+		*ptr = ')';
+		ptr++;
 	}
-	fputc('\n', stdout);
+	*ptr = '\0';
+	df.df_putstr(str, ptr - str);
 }
 
 /*
@@ -739,7 +963,7 @@ static void info_dump(
 	*u_total += cpu_u_usage;
 	*s_total += cpu_s_usage;
 
-	printf("%6.2f %6.2f %6.2f %5d %c %4d %s %s%s%s\n",
+	df.df_printfnl("%6.2f %6.2f %6.2f %5d %c %4d %s %s%s%s",
 		cpu_u_usage + cpu_s_usage,
 		cpu_u_usage, cpu_s_usage,
 		info->pid,
@@ -762,7 +986,7 @@ static void info_total_dump(
 	const double s_total)
 {
 	if (opt_flags & OPT_TOTAL)
-		printf("%6.2f %6.2f %6.2f Total\n",
+		df.df_printfnl("%6.2f %6.2f %6.2f Total",
 			u_total + s_total, u_total, s_total);
 }
 
@@ -804,7 +1028,7 @@ static void samples_dump(
 	if (opt_flags & OPT_GRAND_TOTAL) {
 		double cpu_u_total = 0.0, cpu_s_total = 0.0;
 
-		printf("Grand Total (from %" PRIu32 " samples, %.1f seconds):\n",
+		df.df_printfnl("Grand Total (from %" PRIu32 " samples, %.1f seconds):",
 			samples, duration);
 		info_banner_dump(time_now);
 		for (i = 0; i < n; i++) {
@@ -814,7 +1038,7 @@ static void samples_dump(
 				&cpu_u_total, &cpu_s_total);
 		}
 		info_total_dump(cpu_u_total, cpu_s_total);
-		putchar('\n');
+		df.df_printfnl("");
 	}
 
 	if (!filename) {
@@ -984,7 +1208,7 @@ static void samples_distribution(const uint64_t nr_ticks)
 			bucket[i],
 			100.0 * (double)bucket[i] / (double)total);
 	}
-	putchar('\n');
+	putc('\n', stdout);
 }
 
 /*
@@ -1356,7 +1580,7 @@ static inline void proc_stat_dump(
 	const double duration)
 {
 	double scale = 1.0 / duration;
-	printf("%.1f Ctxt/s, %.1f IRQ/s, %.1f softIRQ/s, "
+	df.df_printfnl("%.1f Ctxt/s, %.1f IRQ/s, %.1f softIRQ/s, "
 		"%.1f new tasks/s, %" PRIu64 " running, %" PRIu64 " blocked\n",
 		scale * delta->ctxt,
 		scale * delta->irq,
@@ -1677,7 +1901,8 @@ static void show_usage(void)
 		" -t specifies a task tick count threshold where samples less\n"
                 "    than this are ignored\n"
 		" -T show total CPU utilisation statistics\n"
-		" -x show extra stats (load average, avg cpu freq, etc)\n");
+		" -x show extra stats (load average, avg cpu freq, etc)\n"
+		" -X top-like curses based display mode\n");
 }
 
 int main(int argc, char **argv)
@@ -1694,12 +1919,13 @@ int main(int argc, char **argv)
 	int32_t n_lines = -1;
 	uint32_t samples = 0;
 	bool forever = true;
+	bool redo = false;
 	int i;
 
 	clock_ticks = (uint64_t)sysconf(_SC_CLK_TCK);
 
 	for (;;) {
-		int c = getopt(argc, argv, "acdDghiln:qr:sSt:Tp:x");
+		int c = getopt(argc, argv, "acdDghiln:qr:sSt:Tp:xX");
 		if (c == -1)
 			break;
 		switch (c) {
@@ -1777,6 +2003,9 @@ int main(int argc, char **argv)
 		case 'x':
 			opt_flags |= OPT_EXTRA_STATS;
 			break;
+		case 'X':
+			opt_flags |= OPT_TOP;
+			break;
 		default:
 			show_usage();
 			exit(EXIT_FAILURE);
@@ -1821,6 +2050,14 @@ int main(int argc, char **argv)
 			exit(EXIT_FAILURE);
 		}
 	}
+	memset(&new_action, 0, sizeof(new_action));
+	new_action.sa_handler = handle_sigwinch;
+	if (sigaction(SIGWINCH, &new_action , NULL) < 0) {
+		fprintf(stderr, "sigaction failed: errno=%d (%s)\n",
+			errno, strerror(errno));
+		exit(EXIT_FAILURE);
+	}
+
 	cpu_stats_old = calloc(TABLE_SIZE, sizeof(cpu_stat_t*));
 	cpu_stats_new = calloc(TABLE_SIZE, sizeof(cpu_stat_t*));
 	if (UNLIKELY(cpu_stats_old == NULL || cpu_stats_new == NULL)) {
@@ -1835,10 +2072,17 @@ int main(int argc, char **argv)
 		get_proc_stat(proc_stat_old);
 	nr_ticks = get_ticks();
 
+	/* Set display functions */
+	df = (opt_flags & OPT_TOP) ? df_top : df_normal;
+	df.df_setup();
+	df.df_winsize(true);
+
 	while (!stop_cpustat && (forever || count--)) {
 		struct timeval tv;
 		double secs, duration, right_now;
 
+		df.df_clear();
+
 		/* Timeout to wait for in the future for this sample */
 		secs = time_start + ((double)t * duration_secs) - time_now;
 		/* Play catch-up, probably been asleep */
@@ -1850,12 +2094,22 @@ int main(int argc, char **argv)
 			if (secs < 0.5)
 				secs += duration_secs;
 		} else {
-			t++;
+			if (!redo)
+				t++;
 		}
+		redo = false;
+
 		double_to_timeval(secs, &tv);
+retry:
 		if (UNLIKELY(select(0, NULL, NULL, NULL, &tv) < 0)) {
 			if (errno == EINTR) {
-				stop_cpustat = true;
+				if (!resized)
+					stop_cpustat = true;
+				else {
+					redo = true;
+					if (timeval_to_double(&tv) > 0.0)
+						goto retry;
+				}
 			} else {
 				fprintf(stderr,
 					"select failed: errno=%d (%s)\n",
@@ -1873,13 +2127,17 @@ int main(int argc, char **argv)
 		total_ticks += nr_ticks;
 		time_now = right_now;
 		get_cpustats(cpu_stats_new, time_now);
+
+		df.df_winsize(redo);
+		df.df_refresh();
+
 		if (opt_flags & OPT_EXTRA_STATS) {
 			double avg_cpu_freq = cpu_freq_average(max_cpus);
 
 			get_proc_stat(proc_stat_new);
 			proc_stat_diff(proc_stat_old, proc_stat_new,
 					&proc_stat_delta);
-			printf("Load Avg %s, Freq Avg. %s, %s CPUs online\n",
+			df.df_printfnl("Load Avg %s, Freq Avg. %s, %s CPUs online",
 				load_average(),
 				cpu_freq_format(avg_cpu_freq),
 				cpus_online());
@@ -1898,9 +2156,14 @@ int main(int argc, char **argv)
 		proc_stat_old = proc_stat_new;
 		proc_stat_new = proc_stat_tmp;
 		samples++;
-		putchar('\n');
+
+		df.df_refresh();
+		df.df_linebreak();
 	}
 
+	df.df_endwin();
+	df = df_normal;
+
 	time_now = gettime_to_double();
 	samples_dump(csv_results, time_now - time_start, time_now,
 		nr_ticks, total_ticks, samples);
